@@ -1,6 +1,6 @@
 /*  _ __   __ _ ___ __  __
 ** | '_ \ / _` |_ _|  \/  | naim
-** | | | | (_| || || |\/| | Copyright 1998-2003 Daniel Reed <n@ml.org>
+** | | | | (_| || || |\/| | Copyright 1998-2005 Daniel Reed <n@ml.org>
 ** |_| |_|\__,_|___|_|  |_| ncurses-based chat client
 */
 #include <naim/naim.h>
@@ -9,44 +9,22 @@
 
 extern win_t	win_info, win_buddy;
 extern conn_t	*curconn;
-extern faimconf_t	faimconf;
+extern faimconf_t faimconf;
 extern time_t	now, awaytime;
 extern double	nowf, changetime;
 extern char	*lastclose;
 extern const char *home;
 extern int	awayc;
 extern awayar_t	*awayar;
-int	buddyc,
-	wbuddy_widthy;
+extern int	printtitle;
 
-void	htmlstrip(char *bb) {
-	char	*start, *end;
-
-	while ((start = strchr(bb, '<')) != NULL)
-		if ((end = strchr(start, '>')) != NULL)
-			memmove(start, end+1, strlen(end));
-		else
-			break;
-	end = bb;
-	while ((start = strchr(end, '&')) != NULL)
-		if ((end = strchr(start, ';')) != NULL) {
-			if (strncasecmp(start+1, "nbsp;", 5) == 0)
-				*start = ' ';
-			else if (strncasecmp(start+1, "amp;", 4) == 0)
-				*start = '&';
-			else if (strncasecmp(start+1, "gt;", 3) == 0)
-				*start = '>';
-			else if (strncasecmp(start+1, "lt;", 3) == 0)
-				*start = '<';
-			else if (strncasecmp(start+1, "quot;", 5) == 0)
-				*start = '"';
-			else
-				continue;
-			memmove(start+1, end+1, strlen(end));
-			end = start+1;
-		} else
-			break;
-}
+extern int
+	buddyc G_GNUC_INTERNAL,
+	wbuddy_widthy G_GNUC_INTERNAL,
+	inplayback G_GNUC_INTERNAL;
+int	buddyc = -1,
+	wbuddy_widthy = -1,
+	inplayback = 0;
 
 static void
 	iupdate(void) {
@@ -88,17 +66,12 @@ static void
 		} else
 			secs_setvar("iftopic", "");
 
+		if ((curconn->curbwin->et != BUDDY) || (curconn->curbwin->e.buddy->typing == 0))
+			secs_setvar("iftyping", "");
+		else
+			secs_setvar("iftyping", secs_script_expand(NULL, getvar(curconn, "statusbar_typing")));
+
 		switch (curconn->curbwin->et) {
-		  case CHAT:
-			if (curconn->curbwin->e.chat->isoper)
-				secs_setvar("ifoper",
-					secs_script_expand(NULL, getvar(curconn, "statusbar_oper")));
-			else
-				secs_setvar("ifoper", "");
-			secs_setvar("ifquery", "");
-			secs_setvar("ifchat",
-				secs_script_expand(NULL, getvar(curconn, "statusbar_chat")));
-			break;
 		  case BUDDY:
 			if (curconn->curbwin->e.buddy->crypt != NULL)
 				secs_setvar("ifcrypt", getvar(curconn, "statusbar_crypt"));
@@ -116,6 +89,16 @@ static void
 				secs_script_expand(NULL, getvar(curconn, "statusbar_query")));
 			secs_setvar("ifchat", "");
 			break;
+		  case CHAT:
+			if (curconn->curbwin->e.chat->isoper)
+				secs_setvar("ifoper",
+					secs_script_expand(NULL, getvar(curconn, "statusbar_oper")));
+			else
+				secs_setvar("ifoper", "");
+			secs_setvar("ifquery", "");
+			secs_setvar("ifchat",
+				secs_script_expand(NULL, getvar(curconn, "statusbar_chat")));
+			break;
 		  case TRANSFER:
 			secs_setvar("ifoper", "");
 			secs_setvar("ifquery", "");
@@ -126,6 +109,7 @@ static void
 		secs_setvar("cur", "");
 		secs_setvar("ifoper", "");
 		secs_setvar("ifquery", "");
+		secs_setvar("iftyping", "");
 		secs_setvar("ifchat", "");
 	}
 
@@ -155,6 +139,16 @@ static void
 		secs_setvar("iflag", "");
 	}
 
+	if (curconn->warnval > 0) {
+		snprintf(buf, sizeof(buf), "%li", curconn->warnval);
+		secs_setvar("warnval", buf);
+		secs_setvar("ifwarn",
+			secs_script_expand(NULL, getvar(curconn, "statusbar_warn")));
+	} else {
+		secs_setvar("warnval", "0");
+		secs_setvar("ifwarn", "");
+	}
+
 	if (strftime(buf, sizeof(buf), getvar(curconn, "statusbar"), tmptr) > 0) {
 		char	*left, *right;
 
@@ -187,6 +181,7 @@ static void
 		nw_printf(&win_info, CB(STATUSBAR,INPUT), 0, "%-*s", faimconf.winfo.widthx,
 			buf);
 		free(left);
+		left = NULL;
 	} else
 		status_echof(curconn, "Error in strftime(): %s.\n", strerror(errno));
 }
@@ -326,6 +321,8 @@ void	bupdate(void) {
 		if ((autoclose > 0) && (curconn->curbwin->et == BUDDY) && !USER_PERMANENT(curconn->curbwin->e.buddy) && (curconn->curbwin->waiting != 0))
 			curconn->curbwin->closetime = now + 60*autoclose;
 		curconn->curbwin->waiting = 0;
+		if ((curconn->curbwin->et == CHAT) && curconn->curbwin->e.chat->isaddressed)
+			curconn->curbwin->e.chat->isaddressed = 0;
 	}
 
 	do {
@@ -345,7 +342,7 @@ void	bupdate(void) {
 			nw_hline(&win_buddy, ACS_HLINE_C | A_BOLD | COLOR_PAIR(C(WINLIST,TEXT)%COLOR_PAIRS), widthx);
 
 			nw_move(&win_buddy, line++, widthx-strlen(conn->winname));
-			if (conn->online == 0) {
+			if (conn->online <= 0) {
 				nw_printf(&win_buddy, C(WINLIST,BUDDY_OFFLINE), 1, " %s", conn->winname);
 				if (line < wbuddy_widthy) {
 					nw_move(&win_buddy, line++, 1);
@@ -397,8 +394,8 @@ void	bupdate(void) {
 					}
 				}
 
-				if (bwin->waiting && (waiting == 0)) {
-					char	tmp[1024];
+				if (bwin->waiting && !waiting) {
+					char	tmp[64];
 
 					if (conn == curconn)
 						snprintf(tmp, sizeof(tmp),
@@ -409,6 +406,10 @@ void	bupdate(void) {
 					secs_setvar("ifpending", tmp);
 					waiting = 1;
 				}
+				if (printtitle && bwin->waiting && (waiting < 2) && ((bwin->et == BUDDY) || ((bwin->et == CHAT) && bwin->e.chat->isaddressed))) {
+					nw_titlef("[%s:%s]", conn->winname, buf);
+					waiting = 2;
+				}
 
 				if (line >= wbuddy_widthy)
 					continue;
@@ -418,19 +419,18 @@ void	bupdate(void) {
 					buf[M] = 0;
 				}
 
-				if (bwin->waiting)
-					col = C(WINLIST,BUDDY_WAITING);
+				if ((bwin->et == CHAT) && bwin->e.chat->isaddressed) {
+					assert(bwin->waiting);
+					col = C(WINLIST,BUDDY_ADDRESSED);
+				} else if (bwin->waiting)
+					if (bwin->et == BUDDY)
+						col = C(WINLIST,BUDDY_ADDRESSED);
+					else
+						col = C(WINLIST,BUDDY_WAITING);
 				else if (bwin->pouncec > 0)
 					col = C(WINLIST,BUDDY_QUEUED);
 				else
 					switch (bwin->et) {
-					  case CHAT:
-						assert(bwin->e.chat != NULL);
-						if (bwin->e.chat->offline)
-							col = C(WINLIST,BUDDY_OFFLINE);
-						else
-							col = C(WINLIST,BUDDY);
-						break;
 					  case BUDDY:
 						if (bwin->e.buddy->tag != NULL)
 							col = CI(WINLIST,BUDDY_TAGGED);
@@ -443,11 +443,35 @@ void	bupdate(void) {
 						else
 							col = C(WINLIST,BUDDY);
 						break;
+					  case CHAT:
+						assert(bwin->e.chat != NULL);
+						if (bwin->e.chat->offline)
+							col = C(WINLIST,BUDDY_OFFLINE);
+						else
+							col = C(WINLIST,BUDDY);
+						break;
 					  case TRANSFER:
 						col = C(WINLIST,BUDDY);
 						break;
 					}
 				assert(col != -1);
+				if (bwin == curconn->curbwin) {
+					int	affect = col/COLOR_PAIRS,
+						back;
+
+#if 0
+					if (faimconf.b[cIMWIN] == faimconf.b[cWINLIST])
+						back = (faimconf.b[cIMWIN]+1)%nw_COLORS;
+					else
+						back = faimconf.b[cIMWIN];
+#else
+						back = faimconf.b[cWINLISTHIGHLIGHT];
+#endif
+
+					col %= nw_COLORS;
+					col += nw_COLORS*back;
+					col += affect*COLOR_PAIRS;
+				}
 				nw_move(&win_buddy, line, 1);
 				if ((col >= 2*COLOR_PAIRS) || (col < COLOR_PAIRS))
 					nw_printf(&win_buddy, col%COLOR_PAIRS, 1, "%*s", M, buf);
@@ -463,11 +487,11 @@ void	bupdate(void) {
 			assert(buddyc < 1000);
 		} while ((bwin = bwin->next) != conn->curbwin);
 
-		if (autosort == 2)
+		if (autosort == 2) {
 			free(lastgroup);
-		else {
+			lastgroup = NULL;
+		} else
 			assert(lastgroup == NULL);
-		}
 	} while ((conn = conn->next) != curconn);
 
 	nw_move(&win_buddy, line-1, 0);
@@ -476,6 +500,8 @@ void	bupdate(void) {
 	else
 		nw_addch(&win_buddy, ACS_HLINE_C | A_BOLD | COLOR_PAIR(C(WINLIST,TEXT)%COLOR_PAIRS));
 
+	if (printtitle && (waiting != 2))
+		nw_titlef("");
 	if (waiting)
 		buddyc = -buddyc;
 	else
@@ -512,8 +538,8 @@ buddywin_t
 		if (bwin->et == et) {
 			if (firetalk_compare_nicks(conn->conn, buddy, bwin->winname) == FE_SUCCESS)
 				return(bwin);
-			if ((bwin->et == BUDDY) && (firetalk_compare_nicks(conn->conn, buddy, USER_NAME(bwin->e.buddy)) == FE_SUCCESS))
-				return(bwin);
+//			if ((bwin->et == BUDDY) && (firetalk_compare_nicks(conn->conn, buddy, USER_NAME(bwin->e.buddy)) == FE_SUCCESS))
+//				return(bwin);
 		}
 	} while ((bwin = bwin->next) != conn->curbwin);
 
@@ -543,11 +569,16 @@ static void
 
 	assert(bwin != NULL);
 	nw_delwin(&(bwin->nwin));
-	for (i = 0; i < bwin->pouncec; i++)
+	for (i = 0; i < bwin->pouncec; i++) {
 		free(bwin->pouncear[i]);
+		bwin->pouncear[i] = NULL;
+	}
 	free(bwin->pouncear);
+	bwin->pouncear = NULL;
 	free(bwin->winname);
+	bwin->winname = NULL;
 	free(bwin->blurb);
+	bwin->blurb = NULL;
 
 	if (bwin->nwin.logfile != NULL) {
 		struct tm	*tmptr;
@@ -556,22 +587,45 @@ static void
 		fprintf(bwin->nwin.logfile, "<I>-----</I> <font color=\"#FFFFFF\">Log file closed %04i-%02i-%02iT%02i:%02i</font> <I>-----</I><br>\n",
 			1900+tmptr->tm_year, 1+tmptr->tm_mon, tmptr->tm_mday, tmptr->tm_hour, tmptr->tm_min);
 		fclose(bwin->nwin.logfile);
+		bwin->nwin.logfile = NULL;
 	}
 
 	free(bwin);
 }
 
+void	verify_winlist_sanity(conn_t *const conn, const buddywin_t *const verifywin) {
+	buddywin_t	*bwin;
+	int	i = 0, found = 0;
+
+	assert(conn != NULL);
+	assert(conn->curbwin != NULL);
+	bwin = conn->curbwin;
+	do {
+		if (bwin == verifywin)
+			found = 1;
+		assert((bwin->et == CHAT) || (bwin->et == BUDDY) || (bwin->et == TRANSFER));
+		if (bwin->et != BUDDY) {
+			assert(bwin->informed == 0);
+			assert(bwin->closetime == 0);
+		}
+		if (bwin->et == CHAT)
+			assert(bwin->keepafterso == 1);
+		assert(strlen(bwin->winname) > 0);
+		assert(i++ < 10000);
+	} while ((bwin = bwin->next) != conn->curbwin);
+	if (verifywin != NULL)
+		assert(found == 1);
+	else
+		assert(found == 0);
+}
+
 void	bclose(conn_t *conn, buddywin_t *bwin, int _auto) {
 	if (bwin == NULL)
 		return;
 
+	verify_winlist_sanity(conn, bwin);
+
 	switch (bwin->et) {
-	  case CHAT:
-		if (bwin->winname[0] != ':')
-			firetalk_chat_part(conn->conn, bwin->winname);
-		free(bwin->e.chat);
-		bwin->e.chat = NULL;
-		break;
 	  case BUDDY:
 		if (_auto == 0) {
 			assert(bwin->e.buddy != NULL);
@@ -580,39 +634,61 @@ void	bclose(conn_t *conn, buddywin_t *bwin, int _auto) {
 			STRREPLACE(lastclose, bwin->winname);
 		}
 		break;
+	  case CHAT:
+		if (bwin->winname[0] != ':')
+			firetalk_chat_part(conn->conn, bwin->winname);
+		free(bwin->e.chat->key);
+		bwin->e.chat->key = NULL;
+		free(bwin->e.chat->last.line);
+		bwin->e.chat->last.line = NULL;
+		free(bwin->e.chat->last.name);
+		bwin->e.chat->last.name = NULL;
+		free(bwin->e.chat);
+		bwin->e.chat = NULL;
+		break;
 	  case TRANSFER:
 		assert(bwin->e.transfer != NULL);
 		firetalk_file_cancel(conn->conn, bwin->e.transfer->handle);
 		echof(conn, NULL, "File transfer aborted.\n");
 		fremove(bwin->e.transfer);
+		bwin->e.transfer = NULL;
 		break;
 	}
 
 	if (bwin == bwin->next)
 		conn->curbwin = NULL;
 	else {
-		buddywin_t	*bbefore;
+		int	i = 0;
+		buddywin_t *bbefore;
 
 		bbefore = bwin->next;
-		while (bbefore->next != bwin)
+		while (bbefore->next != bwin) {
 			bbefore = bbefore->next;
+			assert(i++ < 10000);
+		}
 		bbefore->next = bwin->next;
 
 		if (bwin == conn->curbwin)
 			conn->curbwin = bwin->next;
 	}
 
+	if (conn->curbwin != NULL)
+		verify_winlist_sanity(conn, NULL);
+
 	bremove(bwin);
+	bwin = NULL;
 	bupdate();
 
+	if (conn->curbwin != NULL)
+		verify_winlist_sanity(conn, NULL);
+
 	if (conn->curbwin != NULL)
 		nw_touchwin(&(conn->curbwin->nwin));
 	else
 		nw_touchwin(&(conn->nwin));
 }
 
-const unsigned char *const
-	naim_normalize(const unsigned char *const name) {
+const unsigned char *naim_normalize(const unsigned char *const name) {
 	static char	newname[2048];
 	int	i, j = 0;
 
@@ -686,45 +762,64 @@ static FILE
 
 	if ((rfile = fopen(n, "r")) != NULL) {
 		fclose(rfile);
-		rename(n, nhtml);
+		if ((rfile = fopen(nhtml, "r")) != NULL) {
+			fclose(rfile);
+			status_echof(conn, "Warning: While opening logfile for %s, two versions were found: [%s] and [%s]. I will use [%s], but you may want to look into this discrepency.\n",
+				bwin->winname, n, nhtml, nhtml);
+		} else
+			rename(n, nhtml);
 	}
 
 	return(fopen(nhtml, mode));
 }
 
-void	playback(conn_t *const conn, buddywin_t *const bwin) {
+void	playback(conn_t *const conn, buddywin_t *const bwin, const int lines) {
 	FILE	*rfile;
 
-	assert (bwin->nwin.logfile != NULL);
+	assert(bwin->nwin.logfile != NULL);
 	fflush(bwin->nwin.logfile);
 	bwin->nwin.dirty = 0;
 
 	if ((rfile = playback_fopen(conn, bwin, "r")) != NULL) {
 		char	buf[2048];
-		int	maxlen = faimconf.wstatus.pady*faimconf.wstatus.widthx;
+		int	maxlen = lines*faimconf.wstatus.widthx;
+		long	filesize, playbackstart, playbacklen, pos;
+		time_t	lastprogress = now;
 
 #ifdef DEBUG_ECHO
 		status_echof(conn, "Redrawing window for %s.", bwin->winname);
 #endif
 
-		nw_erase(&win_info);
-		nw_printf(&win_info, CB(CONN,STATUSBAR), 1,
-			" Redrawing window for %s. ",
-			bwin->winname);
-		nw_refresh(&win_info);
+		nw_statusbarf("Redrawing window for %s.", bwin->winname);
 
 		fseek(rfile, 0, SEEK_END);
-		if (ftell(rfile) > maxlen) {
+		while (((filesize = ftell(rfile)) == -1) && (errno == EINTR))
+			;
+		assert(filesize >= 0);
+		if (filesize > maxlen) {
 			fseek(rfile, -maxlen, SEEK_CUR);
 			while ((fgetc(rfile) != '\n') && !feof(rfile))
 				;
 		} else
 			fseek(rfile, 0, SEEK_SET);
+		while (((playbackstart = ftell(rfile)) == -1) && (errno == EINTR))
+			;
+		assert(playbackstart >= 0);
+		pos = 0;
+		playbacklen = filesize-playbackstart;
+		inplayback = 1;
 		while (fgets(buf, sizeof(buf), rfile) != NULL) {
-			while ((strlen(buf) > 0) && (buf[strlen(buf)-1] == '\n'))
-				buf[strlen(buf)-1] = 0;
+			long	len = strlen(buf);
+
+			pos += len;
 			hwprintf(&(bwin->nwin), -C(IMWIN,TEXT)-1, "%s", buf);
+			if ((now = time(NULL)) > lastprogress) {
+				nw_statusbarf("Redrawing window for %s (%li lines left).",
+					bwin->winname, lines*(playbacklen-pos)/playbacklen);
+				lastprogress = now;
+			}
 		}
+		inplayback = 0;
 		fclose(rfile);
 	}
 }
@@ -752,15 +847,15 @@ void	bnewwin(conn_t *conn, const char *name, et_t et) {
 	bwin->keepafterso = bwin->waiting = 0;
 	bwin->et = et;
 	switch (et) {
-	  case CHAT:
-		bwin->e.chat = calloc(1, sizeof(chatlist_t));
-		assert(bwin->e.chat != NULL);
-		bwin->e.chat->offline = 1;
-		break;
 	  case BUDDY:
 		bwin->e.buddy = rgetlist(conn, name);
 		assert(bwin->e.buddy != NULL);
 		break;
+	  case CHAT:
+		bwin->e.chat = calloc(1, sizeof(*(bwin->e.chat)));
+		assert(bwin->e.chat != NULL);
+		bwin->e.chat->offline = 1;
+		break;
 	  case TRANSFER:
 		break;
 	}
@@ -805,12 +900,13 @@ void	bnewwin(conn_t *conn, const char *name, et_t et) {
 				fprintf(bwin->nwin.logfile, "<I>*****</I> <font color=\"#808080\">If you would like help, first try using naim's online help by typing <font color=\"#00FF00\">/help</font>. If you need further help, feel free to ask your question here, and Mr. Reed will get back to you as soon as possible.</font><br>\n");
 				fprintf(bwin->nwin.logfile, "<I>*****</I> <font color=\"#800000\">If you are using Windows telnet to connect to a shell account to run naim, you may notice severe screen corruption. You may wish to try PuTTy, available for free from www.tucows.com. PuTTy handles both telnet and SSH.</font><br>\n");
 			}
+			nw_resize(&(bwin->nwin), 1, 1);
 			bwin->nwin.dirty = 1;
 		}
 	}
 }
 
-void	bcoming(conn_t *conn, const char *buddy, int isaway, int isidle) {
+void	bcoming(conn_t *conn, const char *buddy) {
 	buddywin_t	*bwin = NULL;
 	buddylist_t	*blist = NULL;
 
@@ -828,10 +924,10 @@ void	bcoming(conn_t *conn, const char *buddy, int isaway, int isidle) {
 			assert(bwin != NULL);
 		}
 	}
+	assert((bwin == NULL) || (bwin->e.buddy == blist));
+
 	if (blist->offline == 1) {
-		blist->offline = 0;
-		blist->isaway = isaway;
-		blist->isidle = isidle;
+		blist->isidle = blist->isaway = blist->offline = 0;
 		status_echof(conn, "<font color=\"#00FFFF\">%s</font> <font color=\"#800000\">[<B>%s</B>]</font> is now online =)\n",
 			user_name(NULL, 0, conn, blist), USER_GROUP(blist));
 		if (bwin != NULL) {
@@ -860,45 +956,6 @@ void	bcoming(conn_t *conn, const char *buddy, int isaway, int isidle) {
 			if ((beeponsignon > 1) || ((awaytime == 0) && (beeponsignon == 1)))
 				beep();
 		}
-	} else if (bwin != NULL) {
-		if ((isaway == 1) && (blist->isaway == 0)) {
-			blist->isaway = 1;
-			if ((conn->online+30) < now) {
-				awayc++;
-				awayar = realloc(awayar, awayc*sizeof(*awayar));
-				awayar[awayc-1].name = strdup(buddy);
-				awayar[awayc-1].gotaway = 0;
-				firetalk_im_get_info(conn->conn, buddy);
-			} else
-				window_echof(bwin, "<font color=\"#00FFFF\">%s</font> is now away.\n",
-					user_name(NULL, 0, conn, blist));
-		} else if ((isaway == 0) && (blist->isaway == 1)) {
-			free(bwin->blurb);
-			bwin->blurb = NULL;
-			blist->isaway = 0;
-			window_echof(bwin, "<font color=\"#00FFFF\">%s</font> is no longer away!\n",
-				user_name(NULL, 0, conn, blist));
-		}
-
-		if ((isidle == 1) && (blist->isidle == 0)) {
-			blist->isidle = 1;
-			window_echof(bwin, "<font color=\"#00FFFF\">%s</font> is now idle.\n",
-				user_name(NULL, 0, conn, blist));
-		} else if ((isidle == 0) && (blist->isidle == 1)) {
-			blist->isidle = 0;
-			window_echof(bwin, "<font color=\"#00FFFF\">%s</font> is no longer idle!\n",
-				user_name(NULL, 0, conn, blist));
-		}
-	} else {
-		if ((isaway == 1) && (blist->isaway == 0))
-			blist->isaway = 1;
-		else if ((isaway == 0) && (blist->isaway == 1))
-			blist->isaway = 0;
-
-		if ((isidle == 1) && (blist->isidle == 0))
-			blist->isidle = 1;
-		else if ((isidle == 0) && (blist->isidle == 1))
-			blist->isidle = 0;
 	}
 	bupdate();
 }
@@ -929,7 +986,7 @@ void	bgoing(conn_t *conn, const char *buddy) {
 		status_echof(conn, "<font color=\"#00FFFF\">%s</font> <font color=\"#800000\">[<B>%s</B>]</font> has just logged off :(\n", 
 			user_name(NULL, 0, conn, blist), USER_GROUP(blist));
 		blist->offline = 1;
-		blist->isidle = blist->isaway = 0;
+		blist->warnval = blist->typing = blist->isidle = blist->isaway = 0;
 	} else
 		return;
 
@@ -954,6 +1011,7 @@ void	bgoing(conn_t *conn, const char *buddy) {
 			} else {
 				/* assert(bwin->waiting == 0); */
 				bclose(conn, bwin, 1);
+				bwin = NULL;
 				if ((autoclose > 0) && !USER_PERMANENT(blist)) {
 					rdelbuddy(conn, buddy);
 					firetalk_im_remove_buddy(conn->conn, buddy);
@@ -965,6 +1023,65 @@ void	bgoing(conn_t *conn, const char *buddy) {
 	} while ((bwin = bwin->next) != conn->curbwin);
 }
 
+void	bidle(conn_t *conn, const char *buddy, int isidle) {
+	buddywin_t	*bwin = NULL;
+	buddylist_t	*blist = NULL;
+
+	assert(buddy != NULL);
+	bwin = bgetwin(conn, buddy, BUDDY);
+	if (bwin == NULL)
+		blist = rgetlist(conn, buddy);
+	else
+		blist = bwin->e.buddy;
+	assert(blist != NULL);
+
+	if (bwin != NULL) {
+		if ((isidle == 1) && (blist->isidle == 0))
+			window_echof(bwin, "<font color=\"#00FFFF\">%s</font> is now idle.\n",
+				user_name(NULL, 0, conn, blist));
+		else if ((isidle == 0) && (blist->isidle == 1))
+			window_echof(bwin, "<font color=\"#00FFFF\">%s</font> is no longer idle!\n",
+				user_name(NULL, 0, conn, blist));
+	}
+
+	blist->isidle = isidle;
+}
+
+void	baway(conn_t *conn, const char *buddy, int isaway) {
+	buddywin_t	*bwin = NULL;
+	buddylist_t	*blist = NULL;
+
+	assert(buddy != NULL);
+	bwin = bgetwin(conn, buddy, BUDDY);
+	if (bwin == NULL)
+		blist = rgetlist(conn, buddy);
+	else
+		blist = bwin->e.buddy;
+	assert(blist != NULL);
+
+	if (bwin != NULL) {
+		if ((isaway == 0) && (bwin->blurb != NULL)) {
+			free(bwin->blurb);
+			bwin->blurb = NULL;
+		}
+		if ((isaway == 1) && (blist->isaway == 0)) {
+			if ((conn->online+30) < now) {
+				awayc++;
+				awayar = realloc(awayar, awayc*sizeof(*awayar));
+				awayar[awayc-1].name = strdup(buddy);
+				awayar[awayc-1].gotaway = 0;
+				firetalk_im_get_info(conn->conn, buddy);
+			} else
+				window_echof(bwin, "<font color=\"#00FFFF\">%s</font> is now away.\n",
+					user_name(NULL, 0, conn, blist));
+		} else if ((isaway == 0) && (blist->isaway == 1))
+			window_echof(bwin, "<font color=\"#00FFFF\">%s</font> is no longer away!\n",
+				user_name(NULL, 0, conn, blist));
+	}
+
+	blist->isaway = isaway;
+}
+
 static void
 	bclearall_bwin(conn_t *conn, buddywin_t *bwin, int force) {
 	if (bwin->blurb != NULL) {
@@ -972,14 +1089,6 @@ static void
 		bwin->blurb = NULL;
 	}
 	switch (bwin->et) {
-	  case CHAT:
-		bwin->e.chat->isoper = 0;
-		if (bwin->e.chat->offline == 0) {
-			bwin->e.chat->offline = 1;
-			window_echof(bwin, "Chat <font color=\"#00FFFF\">%s</font> is no longer available :/\n",
-				bwin->winname);
-		}
-		break;
 	  case BUDDY:
 		assert(bwin->e.buddy != NULL);
 		if (bwin->e.buddy->offline == 0) {
@@ -989,14 +1098,25 @@ static void
 		}
 		if (bwin->keepafterso == 0) {
 			bclose(conn, bwin, 1);
+			bwin = NULL;
 			return;
 		}
 		break;
+	  case CHAT:
+		bwin->e.chat->isoper = 0;
+		if (bwin->e.chat->offline == 0) {
+			bwin->e.chat->offline = 1;
+			window_echof(bwin, "Chat <font color=\"#00FFFF\">%s</font> is no longer available :/\n",
+				bwin->winname);
+		}
+		break;
 	  case TRANSFER:
 		break;
 	}
-	if (force)
+	if (force) {
 		bclose(conn, bwin, 1);
+		bwin = NULL;
+	}
 }
 
 static void
@@ -1035,9 +1155,19 @@ void	bclearall(conn_t *conn, int force) {
 	if (conn->buddyar != NULL) {
 		buddylist_t	*blist = conn->buddyar;
 
-		do {
-			bclearall_buddy(blist);
-		} while ((blist = blist->next) != NULL);
+		if (force) {
+			buddylist_t *bnext;
+
+			do {
+				bnext = blist->next;
+				firetalk_im_remove_buddy(conn->conn, blist->_account);
+				do_delbuddy(blist);
+			} while ((blist = bnext) != NULL);
+			conn->buddyar = NULL;
+		} else
+			do {
+				bclearall_buddy(blist);
+			} while ((blist = blist->next) != NULL);
 	}
 
 	bupdate();
