@@ -1,6 +1,6 @@
 /*  _ __   __ _ ___ __  __
 ** | '_ \ / _` |_ _|  \/  | naim
-** | | | | (_| || || |\/| | Copyright 1998-2004 Daniel Reed <n@ml.org>
+** | | | | (_| || || |\/| | Copyright 1998-2006 Daniel Reed <n@ml.org>
 ** |_| |_|\__,_|___|_|  |_| ncurses-based chat client
 */
 #include <naim/naim.h>
@@ -16,15 +16,14 @@ extern win_t	win_input;
 extern conn_t	*curconn;
 extern time_t	now, awaytime;
 extern double	nowf;
-extern char	*namesbuf;
+extern int	namec;
+extern char	**names;
 extern namescomplete_t	namescomplete;
 extern faimconf_t	faimconf;
-extern char	*sty;
+extern char	*sty, *statusbar_text;
 
-extern int
-	awayc G_GNUC_INTERNAL;
-extern awayar_t
-	*awayar G_GNUC_INTERNAL;
+extern int awayc G_GNUC_INTERNAL;
+extern awayar_t *awayar G_GNUC_INTERNAL;
 int	awayc = 0;
 awayar_t	*awayar = NULL;
 
@@ -80,12 +79,10 @@ static void str_replace(const unsigned char *orig, const unsigned char *new, uns
 	}
 }
 
-html_clean_t
-	*html_cleanar = NULL;
+html_clean_t *html_cleanar = NULL;
 int	html_cleanc = 0;
 
-static const unsigned char
-	*html_clean(const unsigned char *str) {
+static const unsigned char *html_clean(const unsigned char *str) {
 	static unsigned char
 		buf[1024*4];
 	int	i;
@@ -100,26 +97,6 @@ static const unsigned char
 	return(buf);
 }
 
-int	getvar_int(conn_t *conn, const char *str) {
-	char	*ptr;
-	char	buf[1024];
-
-	snprintf(buf, sizeof(buf), "%s:%s", conn->winname, str);
-	if ((ptr = secs_getvar(buf)) != NULL)
-		return(atoi(ptr));
-	return(secs_getvar_int(str));
-}
-
-char	*getvar(conn_t *conn, const char *str) {
-	char	buf[1024];
-	char	*val;
-
-	snprintf(buf, sizeof(buf), "%s:%s", conn->winname, str);
-	if ((val = secs_getvar(buf)) != NULL)
-		return(val);
-	return(secs_getvar(str));
-}
-
 nFIRE_HANDLER(naim_newnick) {
 	conn_t		*conn = (conn_t *)client;
 	va_list		msg;
@@ -165,19 +142,26 @@ nFIRE_HANDLER(naim_nickchange) {
 nFIRE_HANDLER(naim_buddylist) {
 	conn_t		*conn = (conn_t *)client;
 	va_list		msg;
-	const char	*screenname, *group;
+	const char	*screenname, *group, *friendly;
 	int		online, away;
 	buddylist_t	*blist;
 
 	va_start(msg, client);
 	screenname = va_arg(msg, const char *);
 	group = va_arg(msg, const char *);
+	friendly = va_arg(msg, const char *);
 	online = va_arg(msg, int);
 	away = va_arg(msg, int);
 	va_end(msg);
 
-	if (rgetlist(conn, screenname) == NULL)
-		blist = raddbuddy(conn, screenname, group, NULL);
+	if ((blist = rgetlist(conn, screenname)) == NULL)
+		blist = raddbuddy(conn, screenname, group, friendly);
+
+	STRREPLACE(blist->_group, group);
+	if (friendly != NULL)
+		STRREPLACE(blist->_name, friendly);
+	else
+		FREESTR(blist->_name);
 }
 
 #define STANDARD_TRAILER	\
@@ -186,15 +170,9 @@ nFIRE_HANDLER(naim_buddylist) {
 
 void	naim_set_info(void *sess, const char *str) {
 	const char *defar[] = {
-"Common sense and a sense of humor are the same thing, moving at\n"
-"different speeds. A sense of humor is just common sense, dancing.<br>\n"
-"&nbsp; &nbsp; &nbsp; &nbsp; -- Clive James<br>\n"
+"Tepid! Tepid is no good for a star, but it'll do for stardust.<br>\n"
 			STANDARD_TRAILER,
-"If you pick up a starving dog and make him prosperous, he will not bite\n"
-"you. This is the principal difference between a dog and a man.<br>\n"
-"&nbsp; &nbsp; &nbsp; &nbsp; -- Mark Twain<br>\n"
-			STANDARD_TRAILER,
-"Resisting temptation is easier when you think you'll probably get\n"
+"Resisting temptation is easier when you think you'll probably get "
 "another chance later on.<br>\n"
 			STANDARD_TRAILER,
 "A little inaccuracy sometimes saves tons of explanation.<br>\n"
@@ -204,17 +182,23 @@ void	naim_set_info(void *sess, const char *str) {
 			STANDARD_TRAILER,
 "<a href=\"http://creativecommons.org/\">Creativity always builds on the past.</a><br>\n"
 			STANDARD_TRAILER,
+"Those who make peaceful revolution impossible will make violent "
+"revolution inevitable.<br>\n"
+"&nbsp; &nbsp; &nbsp; &nbsp; -- John F. Kennedy<br>\n"
+			STANDARD_TRAILER,
+"The question of whether computers can think is just like the question of whether submarines can swim.<br>\n"
+"&nbsp; &nbsp; &nbsp; &nbsp; -- Edsger W. Dijkstra<br>\n"
+			STANDARD_TRAILER,
+"People sleep peaceably in their beds at night only because "
+"rough men stand ready to do violence on their behalf.<br>\n"
+			STANDARD_TRAILER,
+"If you can't fix it, you gotta stand it.<br>\n"
+			STANDARD_TRAILER,
+"While you don't greatly need the outside world, it's still very "
+"reassuring to know that it's still there.<br>\n"
+			STANDARD_TRAILER,
 	};
 
-	if (awaytime > 0) {
-		char	buf[1024];
-
-		snprintf(buf, sizeof(buf), "%lu :%s", (now-awaytime)/60, 
-			secs_getvar("awaymsg"));
-		firetalk_subcode_send_reply(sess, NULL, "AWAY", buf);
-	} else
-		firetalk_subcode_send_reply(sess, NULL, "AWAY", NULL);
-
 	if (str != NULL)
 		firetalk_set_info(sess, str);
 	else {
@@ -270,11 +254,10 @@ void	naim_setversion(conn_t *conn) {
 		snprintf(naim_version, sizeof(naim_version), "%s:unknown, %ix%i%s%s, %s %s", NAIM_VERSION_STRING,
 			COLS, LINES, where, where2, term, lang);
 
-	firetalk_subcode_send_reply(conn->conn, NULL, "VERSION", 
-		naim_version);
+	firetalk_subcode_register_request_reply(conn->conn, "VERSION", naim_version);
 }
 
-nFIRE_HANDLER(naim_budprof) {
+nFIRE_HANDLER(naim_doinit) {
 	conn_t		*conn = (conn_t *)client;
 	va_list		msg;
 	buddylist_t	*buddy;
@@ -291,21 +274,25 @@ nFIRE_HANDLER(naim_budprof) {
 
 	status_echof(conn, "Uploading buddy/block list and profile...\n");
 
-	for (buddy = conn->buddyar; buddy != NULL; buddy = buddy->next)
-		firetalk_im_add_buddy(sess, USER_ACCOUNT(buddy), USER_GROUP(buddy));
+	for (buddy = conn->buddyar; buddy != NULL; buddy = buddy->next) {
+		fte_t	ret;
+
+		ret = firetalk_im_add_buddy(sess, USER_ACCOUNT(buddy), USER_GROUP(buddy), buddy->_name);
+		assert(ret == FE_SUCCESS);
+	}
 
 	for (idiot = conn->idiotar; idiot != NULL; idiot = idiot->next)
-		if (strcasecmp(idiot->notes, "block") == 0)
-			firetalk_im_internal_add_deny(sess, idiot->screenname);
+		if (strcasecmp(idiot->notes, "block") == 0) {
+			fte_t	ret;
+
+			ret = firetalk_im_internal_add_deny(sess, idiot->screenname);
+			assert(ret == FE_SUCCESS);
+		}
 
 	naim_set_info(sess, conn->profile);
 
-	if (awaytime > 0) {
-		int	ret;
-
-		ret = firetalk_set_away(sess, secs_getvar("awaymsg"), 0);
-		assert(ret != FE_NOTCONNECTED);
-	}
+	if (awaytime > 0)
+		firetalk_set_away(sess, secs_getvar("awaymsg"), 0);
 }
 
 nFIRE_HANDLER(naim_setidle) {
@@ -334,6 +321,7 @@ nFIRE_HANDLER(naim_warned) {
 
 	echof(conn, NULL, "<font color=\"#00FFFF\">%s</font> just warned you (%i).\n",
 		who, newlev);
+	conn->warnval = newlev;
 }
 
 nFIRE_HANDLER(naim_buddy_idle) {
@@ -353,6 +341,82 @@ nFIRE_HANDLER(naim_buddy_idle) {
 		bidle(conn, who, 0);
 }
 
+nFIRE_HANDLER(naim_buddy_eviled) {
+	conn_t		*conn = (conn_t *)client;
+	va_list		msg;
+	const char	*who;
+	long		warnval;
+	buddylist_t	*blist;
+
+	va_start(msg, client);
+	who = va_arg(msg, const char *);
+	warnval = va_arg(msg, long);
+	va_end(msg);
+
+	if ((blist = rgetlist(conn, who)) != NULL)
+		blist->warnval = warnval;
+}
+
+nFIRE_HANDLER(naim_buddy_caps) {
+	conn_t		*conn = (conn_t *)client;
+	va_list		msg;
+	const char	*who, *caps;
+	buddylist_t	*blist;
+
+	va_start(msg, client);
+	who = va_arg(msg, const char *);
+	caps = va_arg(msg, const char *);
+	va_end(msg);
+
+	if ((blist = rgetlist(conn, who)) != NULL) {
+		int	i, j, strtolower = 1;
+
+		blist->caps = realloc(blist->caps, 2*strlen(caps)+1);
+
+		for (i = 0; (caps[i] != 0) && (caps[i] != ' '); i++)
+			if (islower(caps[i])) {
+				strtolower = 0;
+				break;
+			}
+
+		for (j = i = 0; caps[i] != 0; i++)
+			if (caps[i] == ' ') {
+				int	x;
+
+				strtolower = 1;
+				for (x = i+1; (caps[x] != 0) && (caps[x] != ' '); x++)
+					if (islower(caps[x])) {
+						strtolower = 0;
+						break;
+					}
+				blist->caps[j++] = ',';
+				blist->caps[j++] = ' ';
+			} else if (caps[i] == '_')
+				blist->caps[j++] = ' ';
+			else if (strtolower && (j > 0) && (blist->caps[j-1] != ' '))
+				blist->caps[j++] = tolower(caps[i]);
+			else
+				blist->caps[j++] = caps[i];
+		blist->caps[j] = 0;
+	}
+}
+
+nFIRE_HANDLER(naim_buddy_typing) {
+	conn_t		*conn = (conn_t *)client;
+	va_list		msg;
+	const char	*who;
+	int		typing;
+	buddylist_t	*blist;
+
+	va_start(msg, client);
+	who = va_arg(msg, const char *);
+	typing = va_arg(msg, int);
+	va_end(msg);
+
+	if ((blist = rgetlist(conn, who)) != NULL)
+		blist->typing = typing;
+}
+
 nFIRE_HANDLER(naim_buddy_away) {
 	conn_t		*conn = (conn_t *)client;
 	va_list		msg;
@@ -377,6 +441,59 @@ nFIRE_HANDLER(naim_buddy_unaway) {
 	baway(conn, who, 0);
 }
 
+nFIRE_HANDLER(naim_buddyadded) {
+	conn_t		*conn = (conn_t *)client;
+	va_list		msg;
+	const char	*screenname, *group, *friendly;
+	buddylist_t	*blist;
+
+	va_start(msg, client);
+	screenname = va_arg(msg, const char *);
+	group = va_arg(msg, const char *);
+	friendly = va_arg(msg, const char *);
+	va_end(msg);
+
+	if ((blist = rgetlist(conn, screenname)) == NULL) {
+		if (conn->online > 0)
+			status_echof(conn, "User <font color=\"#00FFFF\">%s</font> was added to your buddy list by another client signed on as you.\n",
+				screenname);
+		blist = raddbuddy(conn, screenname, group, friendly);
+	}
+
+	STRREPLACE(blist->_group, group);
+	if (friendly != NULL)
+		STRREPLACE(blist->_name, friendly);
+	else
+		FREESTR(blist->_name);
+}
+
+nFIRE_HANDLER(naim_buddyremoved) {
+	conn_t		*conn = (conn_t *)client;
+	va_list		msg;
+	const char	*screenname;
+	buddywin_t	*bwin;
+	buddylist_t	*blist;
+
+	va_start(msg, client);
+	screenname = va_arg(msg, const char *);
+	va_end(msg);
+
+	if ((bwin = bgetwin(conn, screenname, BUDDY)) != NULL) {
+		status_echof(conn, "Closed window <font color=\"#00FFFF\">%s</font> due to buddy removal.\n",
+			bwin->winname);
+		bclose(conn, bwin, 0);
+	}
+
+	if ((blist = rgetlist(conn, screenname)) != NULL) {
+		if (conn->online > 0)
+			status_echof(conn, "User <font color=\"#00FFFF\">%s</font> was removed from your buddy list by another client signed on as you.\n",
+				screenname);
+		rdelbuddy(conn, screenname);
+	}
+}
+
+HOOK_DECLARE(proto_user_onlineval);
+
 nFIRE_HANDLER(naim_buddy_coming) {
 	conn_t		*conn = (conn_t *)client;
 	va_list		msg;
@@ -387,6 +504,7 @@ nFIRE_HANDLER(naim_buddy_coming) {
 	va_end(msg);
 
 	bcoming(conn, who);
+	HOOK_CALL(proto_user_onlineval, (client, conn, who, NULL, NULL, 1));
 }
 
 nFIRE_HANDLER(naim_buddy_going) {
@@ -399,13 +517,13 @@ nFIRE_HANDLER(naim_buddy_going) {
 	va_end(msg);
 
 	bgoing(conn, who);
+	HOOK_CALL(proto_user_onlineval, (client, conn, who, NULL, NULL, 0));
 }
 
 
 
 HOOK_DECLARE(recvfrom);
-static void
-	naim_recvfrom(conn_t *const conn,
+static void naim_recvfrom(conn_t *const conn,
 		const char *const _name, 
 		const char *const _dest,
 		const unsigned char *_message, int len,
@@ -426,8 +544,7 @@ static void
 	free(message);
 }
 
-static int
-	recvfrom_ignorelist(conn_t *conn, char **name, char **dest, 
+static int recvfrom_ignorelist(conn_t *conn, char **name, char **dest, 
 		unsigned char **message, int *len, int *flags) {
 	ignorelist_t	*ig;
 
@@ -443,8 +560,7 @@ static int
 	return(HOOK_CONTINUE);
 }
 
-static int
-	recvfrom_decrypt(conn_t *conn, char **name, char **dest,
+static int recvfrom_decrypt(conn_t *conn, char **name, char **dest,
 		unsigned char **message, int *len, int *flags) {
 	if ((*dest == NULL) && !(*flags & RF_ACTION)) {
 		buddylist_t	*blist = rgetlist(conn, *name);
@@ -452,12 +568,6 @@ static int
 		if ((blist != NULL) && (blist->crypt != NULL) && ((blist->peer <= 3) || (*flags & RF_ENCRYPTED))) {
 			int	i, j = 0;
 
-#if 0
-			echof(curconn, "DECRYPT", "<-- %s = %s", *name, *message);
-			for (i = 0; blist->crypt[i] != 0; i++)
-				echof(curconn, "DECRYPT", "%i: %c ^ %c = %02X", i, (*message)[i] ^ (unsigned char)blist->crypt[i], (unsigned char)blist->crypt[i], (*message)[i]);
-#endif
-
 			for (i = 0; i < *len; i++) {
 				(*message)[i] = (*message)[i] ^ (unsigned char)blist->crypt[j++];
 				if (blist->crypt[j] == 0)
@@ -468,25 +578,19 @@ static int
 					*len, i, (*message)[i]);
 				return(HOOK_STOP);
 			}
-
-#if 0
-			echof(curconn, "DECRYPT", "<-- %s %s ^ %s", *name, blist->crypt, *message);
-#endif
 		}
 	}
 	return(HOOK_CONTINUE);
 }
 
-static int
-	recvfrom_log(conn_t *conn, char **name, char **dest,
+static int recvfrom_log(conn_t *conn, char **name, char **dest,
 		unsigned char **message, int *len, int *flags) {
 	if (!(*flags & RF_NOLOG))
 		logim(conn, *name, *dest, *message);
 	return(HOOK_CONTINUE);
 }
 
-static int
-	recvfrom_beep(conn_t *conn, char **name, char **dest,
+static int recvfrom_beep(conn_t *conn, char **name, char **dest,
 		unsigned char **message, int *len, int *flags) {
 	if (*dest == NULL) {
 		int	beeponim = getvar_int(conn, "beeponim");
@@ -497,8 +601,7 @@ static int
 	return(HOOK_CONTINUE);
 }
 
-static int
-	recvfrom_autobuddy(conn_t *conn, char **name, char **dest,
+static int recvfrom_autobuddy(conn_t *conn, char **name, char **dest,
 		unsigned char **message, int *len, int *flags) {
 	if (*dest == NULL) {
 		buddylist_t	*blist = rgetlist(conn, *name);
@@ -510,7 +613,7 @@ static int
 				blist = raddbuddy(conn, *name, DEFAULT_GROUP, NULL);
 				assert(blist->offline == 1);
 				bnewwin(conn, *name, BUDDY);
-				firetalk_im_add_buddy(conn->conn, *name, USER_GROUP(blist));
+				firetalk_im_add_buddy(conn->conn, *name, USER_GROUP(blist), NULL);
 			} else {
 				buddywin_t	*bwin;
 
@@ -524,8 +627,7 @@ static int
 	return(HOOK_CONTINUE);
 }
 
-static int
-	recvfrom_display_user(conn_t *conn, char **name, char **dest,
+static int recvfrom_display_user(conn_t *conn, char **name, char **dest,
 		unsigned char **message, int *len, int *flags) {
 	buddylist_t	*blist;
 	buddywin_t	*bwin;
@@ -541,7 +643,7 @@ static int
 
 		naim_lastupdate(conn);
 
-		if (*flags & RF_AUTOMATIC) {
+		if (*flags & RF_ACTION) {
 			if ((blist != NULL) && (blist->crypt != NULL))
 				format = "<B>*&gt; %s</B>";
 			else
@@ -563,7 +665,7 @@ static int
 		** window, and we want them to remain unambiguous.
 		*/
 		hwprintf(&(conn->nwin), C(CONN,BUDDY), format, *name);
-		hwprintf(&(conn->nwin), C(CONN,TEXT), " <body>%s</body><br>", *message);
+		hwprintf(&(conn->nwin), C(CONN,TEXT), "%s<body>%s</body><br>", (strncmp(*message, "'s ", 3) == 0)?"":" ", *message);
 	} else {
 		const char *format;
 
@@ -594,7 +696,7 @@ static int
 
 		WINTIME(&(bwin->nwin), IMWIN);
 		hwprintf(&(bwin->nwin), C(IMWIN,BUDDY), format, USER_NAME(blist));
-		hwprintf(&(bwin->nwin), C(IMWIN,TEXT), " <body>%s</body><br>", *message);
+		hwprintf(&(bwin->nwin), C(IMWIN,TEXT), "%s<body>%s</body><br>", (strncmp(*message, "'s ", 3) == 0)?"":" ", *message);
 
 		if (!(*flags & RF_AUTOMATIC)) {
 			int	autoreply = getvar_int(conn, "autoreply");
@@ -616,12 +718,50 @@ static int
 	return(HOOK_CONTINUE);
 }
 
-static int
-	recvfrom_display_chat(conn_t *conn, char **name, char **dest,
+static void recvfrom_display_chat_print(buddywin_t *bwin, const int flags, const int istome, const char *name, const char *prefix, const unsigned char *message) {
+	const char	*format;
+
+	if (prefix == NULL)
+		prefix = "&gt;";
+
+	if (flags & RF_ACTION)
+		format = "* <B>%s</B>";
+	else if (flags & RF_AUTOMATIC)
+		format = "-<B>%s</B>-";
+	else
+		format = "&lt;<B>%s</B>%s";
+	hwprintf(&(bwin->nwin), istome?C(IMWIN,BUDDY_ADDRESSED):C(IMWIN,BUDDY), 
+		format, name, prefix);
+
+	hwprintf(&(bwin->nwin), C(IMWIN,TEXT), " <body>%s%s%s</body><br>",
+		istome?"<B>":"", message, istome?"</B>":"");
+}
+
+void	chat_flush(buddywin_t *bwin) {
+	assert(bwin->et == CHAT);
+	assert(bwin->e.chat->last.reps >= 0);
+	if (bwin->e.chat->last.reps > 0) {
+		if (bwin->e.chat->last.reps == 1) {
+			assert(bwin->e.chat->last.lasttime != 0);
+			WINTIME_NOTNOW(&(bwin->nwin), IMWIN, bwin->e.chat->last.lasttime);
+			recvfrom_display_chat_print(bwin, bwin->e.chat->last.flags, bwin->e.chat->last.istome, bwin->e.chat->last.name, NULL, bwin->e.chat->last.line);
+		} else {
+			assert(bwin->e.chat->last.lasttime != 0);
+			WINTIME_NOTNOW(&(bwin->nwin), IMWIN, bwin->e.chat->last.lasttime);
+			hwprintf(&(bwin->nwin), C(IMWIN,TEXT), "<B>[Last message repeated %i more times]</B><br>", bwin->e.chat->last.reps);
+		}
+		bwin->e.chat->last.reps = 0;
+	}
+	FREESTR(bwin->e.chat->last.line);
+	FREESTR(bwin->e.chat->last.name);
+}
+
+static int recvfrom_display_chat(conn_t *conn, char **name, char **dest,
 		unsigned char **message, int *len, int *flags) {
 	buddywin_t	*bwin;
 	int		istome;
-	const char	*format;
+	char		*prefix = NULL;
+	unsigned char	*message_save;
 
 	if (*dest == NULL)
 		return(HOOK_CONTINUE);
@@ -629,7 +769,9 @@ static int
 	{
 		unsigned char *match;
 
-		if ((aimncmp(*message, conn->sn, strlen(conn->sn)) == 0) && !isalpha(*(*message+strlen(conn->sn))))
+		if (conn->sn == NULL)
+			istome = 0;
+		else if ((aimncmp(*message, conn->sn, strlen(conn->sn)) == 0) && !isalpha(*(*message+strlen(conn->sn))))
 			istome = 1;
 		else if (((match = strstr(*message, conn->sn)) != NULL)
 			&& ((match == *message) || !isalpha(*(match-1)))
@@ -647,17 +789,64 @@ static int
 			bwin->e.chat->isaddressed = 1;
 	}
 
+	if (bwin->e.chat->last.line != NULL)
+		assert(bwin->e.chat->last.name != NULL);
+	if ((bwin->e.chat->last.line != NULL) && (*flags == bwin->e.chat->last.flags)
+		&& (firetalk_compare_nicks(conn->conn, *name, bwin->e.chat->last.name) == FE_SUCCESS)
+		&& (strcasecmp(*message, bwin->e.chat->last.line) == 0)) {
+		bwin->e.chat->last.reps++;
+		bwin->e.chat->last.lasttime = now;
+		return(HOOK_CONTINUE);
+	}
+	if (bwin->e.chat->last.reps > 0)
+		chat_flush(bwin);
+	assert(bwin->e.chat->last.reps == 0);
+
+	message_save = strdup(*message);
+
+	if ((bwin->winname[0] != ':') && (bwin->e.chat->last.line != NULL)) {
+		char	*tmp = strdup(*message);
+		size_t	off, add;
+
+		htmlreplace(tmp, '_');
+		off = strspn(tmp, "_");
+		add = strcspn(tmp+off, " ");
+		assert(bwin->e.chat->last.name != NULL);
+		if ((add > off) && isalpha(tmp[off]) && !isalpha(tmp[off+add-1])) {
+			if (strncasecmp(tmp+off, bwin->e.chat->last.name, strlen(bwin->e.chat->last.name)) == 0) {
+				static int	sent_carat_desc = 0;
+
+				if (!sent_carat_desc) {
+					statusbar_text = strdup("A ^ near a speaker name indicates that message was addressed to the previous speaker.");
+					sent_carat_desc = 1;
+				}
+				prefix = "^";
+				memmove((*message)+off, (*message)+off+add, strlen((*message)+off+add)+1);
+			} else if (strncasecmp(*message, bwin->e.chat->last.line, off+add) == 0) {
+				static int	sent_plus_desc = 0;
+
+				if (!sent_plus_desc) {
+					statusbar_text = strdup("A + near a speaker name indicates that message was addressed to the same person as the previous message.");
+					sent_plus_desc = 1;
+				}
+				if (firetalk_compare_nicks(conn->conn, *name, bwin->e.chat->last.name) != FE_SUCCESS)
+					prefix = "+";
+				memmove((*message)+off, (*message)+off+add, strlen((*message)+off+add)+1);
+			}
+		}
+		free(tmp);
+	}
+
 	WINTIME(&(bwin->nwin), IMWIN);
-	if (*flags & RF_ACTION)
-		format = "* <B>%s</B>";
-	else if (*flags & RF_AUTOMATIC)
-		format = "-<B>%s</B>-";
-	else
-		format = "&lt;<B>%s</B>&gt;";
-	hwprintf(&(bwin->nwin), istome?C(IMWIN,EVENT):C(IMWIN,BUDDY), 
-		format, *name);
-	hwprintf(&(bwin->nwin), C(IMWIN,TEXT), " <body>%s%s%s</body><br>",
-		istome?"<B>":"", *message, istome?"</B>":"");
+	recvfrom_display_chat_print(bwin, *flags, istome, *name, prefix, *message);
+
+	free(bwin->e.chat->last.line);
+	bwin->e.chat->last.line = message_save;
+	STRREPLACE(bwin->e.chat->last.name, (*name));
+	bwin->e.chat->last.reps = 0;
+	bwin->e.chat->last.lasttime = now;
+	bwin->e.chat->last.flags = *flags;
+	bwin->e.chat->last.istome = istome;
 
 	bupdate();
 	return(HOOK_CONTINUE);
@@ -776,7 +965,7 @@ nFIRE_HANDLER(naim_connected) {
 	conn_t		*conn = (conn_t *)client;
 	buddywin_t	*bwin = conn->curbwin;
 
-	if (conn->online != 0) {
+	if (conn->online > 0) {
 		status_echof(conn, "naim just received notification that you have connected to %s at %lu,"
 			" but I'm pretty sure you've been connected since %lu."
 			" This is a bug, and your session may be unstable.\n",
@@ -828,6 +1017,26 @@ nFIRE_HANDLER(naim_connectfailed) {
 		echof(conn, "CONNECT", "Unable to connect to %s: %s.\n",
 			firetalk_strprotocol(conn->proto),
 			firetalk_strerror(err));
+
+	if (err == FE_BADUSER) {
+		char	*str;
+
+		echof(conn, NULL, "Attempting to reconnect using a different name...\n");
+		str = malloc(strlen(conn->sn)+2);
+		strcpy(str, conn->sn);
+		if (strlen(conn->sn) > 8) {
+			int	pos = rand()%(strlen(conn->sn)-1)+1;
+
+			if (str[pos] == '_')
+				str[pos] = rand()%9+'1';
+			else
+				str[pos] = '_';
+		} else
+			strcat(str, "_");
+		free(conn->sn);
+		conn->sn = str;
+		conio_connect(conn, 0, NULL);
+	}
 }
 
 nFIRE_HANDLER(naim_error_msg) {
@@ -846,6 +1055,32 @@ nFIRE_HANDLER(naim_error_msg) {
 	if ((error == FE_MESSAGETRUNCATED) && (awaytime > 0))
 		return;
 
+	if ((error == FE_INVALIDFORMAT) && (target != NULL) && (awayc > 0)) {
+		int	i;
+
+		assert(awayar != NULL);
+
+		for (i = 0; i < awayc; i++)
+			if (firetalk_compare_nicks(conn->conn, target, awayar[i].name) == FE_SUCCESS) {
+				if (awayar[i].gotaway == 0) {
+					buddywin_t *bwin = bgetwin(conn, target, BUDDY);
+
+					if (bwin == NULL)
+						status_echof(conn, "<font color=\"#00FFFF\">%s</font> is now away.\n",
+							target);
+					else
+						window_echof(bwin, "<font color=\"#00FFFF\">%s</font> is now away.\n",
+							user_name(NULL, 0, conn, bwin->e.buddy));
+				}
+				free(awayar[i].name);
+				memmove(awayar+i, awayar+i+1, (awayc-i-1)*sizeof(*awayar));
+				awayc--;
+				awayar = realloc(awayar, awayc*sizeof(*awayar));
+				return;
+			}
+	}
+
+
 	if ((target != NULL) && ((bwin = bgetanywin(conn, target)) != NULL)) {
 		if (desc != NULL)
 			window_echof(bwin, "ERROR: %s, %c%s\n",
@@ -883,7 +1118,7 @@ nFIRE_HANDLER(naim_error_disconnect) {
 
 	echof(conn, NULL, "Disconnected from %s: %s.\n",
 		conn->winname, firetalk_strerror(error));
-	conn->online = 0;
+	conn->online = -1;
 	bclearall(conn, 0);
 
 	if (error == FE_RECONNECTING)
@@ -975,6 +1210,13 @@ nFIRE_HANDLER(naim_userinfo_handler) {
 			(class&FF_SUBSTANDARD)?" AIM":"",
 			(class&FF_NORMAL)?" AOLamer":"",
 			(class&FF_ADMIN)?" Operator":"");
+	{
+		buddylist_t *blist = rgetlist(conn, SN);
+
+		if ((blist != NULL) && (blist->caps != NULL))
+		  echof(conn, NULL,
+			"</B><B>Client features</B>: %s", blist->caps);
+	}
 	if (warning > 0)
 	  echof(conn, NULL,
 		"</B>&nbsp; <B>Warning level</B>: %i", warning);
@@ -1094,11 +1336,12 @@ nFIRE_HANDLER(naim_chat_JOIN) {
 	conn_t		*conn = (conn_t *)client;
 	va_list		msg;
 	buddywin_t	*bwin;
-	const char	*room, *who, *q;
+	const char	*room, *who, *q, *extra;
 
 	va_start(msg, client);
 	room = va_arg(msg, const char *);
 	who = va_arg(msg, const char *);
+	extra = va_arg(msg, const char *);
 	va_end(msg);
 
 	q = (strchr(room, ' ') != NULL)?"\"":"";
@@ -1108,8 +1351,12 @@ nFIRE_HANDLER(naim_chat_JOIN) {
 	if (getvar_int(conn, "chatverbose") & CH_USERS)
 		bwin->waiting = 1;
 
-	window_echof(bwin, "<font color=\"#00FFFF\">%s</font> has joined chat %s%s%s.\n",
-		who, q, room, q);
+	if (extra == NULL)
+		window_echof(bwin, "<font color=\"#00FFFF\">%s</font> has joined chat %s%s%s.\n",
+			who, q, room, q);
+	else
+		window_echof(bwin, "<font color=\"#00FFFF\">%s</font> (%s) has joined chat %s%s%s.\n",
+			who, extra, q, room, q);
 	bupdate();
 }
 
@@ -1189,8 +1436,7 @@ nFIRE_HANDLER(naim_chat_KEYCHANGED) {
 	} else if (bwin->e.chat->key != NULL) {
 		window_echof(bwin, "<font color=\"#00FFFF\">%s</font> has cleared the channel key (was %s).\n",
 			by, bwin->e.chat->key);
-		free(bwin->e.chat->key);
-		bwin->e.chat->key = NULL;
+		FREESTR(bwin->e.chat->key);
 	}
 	bupdate();
 }
@@ -1353,8 +1599,7 @@ nFIRE_HANDLER(naim_chat_NAMES) {
 	va_list		msg;
 	const char	*room, *nick;
 	int		oped;
-	static int	namesbuflen = 0;
-	int		i, j;
+//	int		i, j;
 
 	va_start(msg, client);
 	room = va_arg(msg, const char *);
@@ -1396,26 +1641,20 @@ nFIRE_HANDLER(naim_chat_NAMES) {
 		return;
 	}
 
-	if (namesbuf == NULL) {
-		namesbuflen = 1;
-		namesbuf = malloc(1);
-		namesbuf[0] = 0;
-	}
-	namesbuflen += strlen(nick) + oped + 1;
-	namesbuf = realloc(namesbuf, namesbuflen);
-	if (oped)
-		strcat(namesbuf, "@");
-	for (i = 0, j = strlen(namesbuf); nick[i] != 0; i++, j++)
-		if (isspace(nick[i]))
-			namesbuf[j] = '_';
-		else
-			namesbuf[j] = nick[i];
-	namesbuf[j] = ' ';
-	namesbuf[j+1] = 0;
+	namec++;
+	names = realloc(names, namec*sizeof(*names));
+	names[namec-1] = malloc(strlen(nick) + oped + 1);
+	sprintf(names[namec-1], "%s%s", oped?"@":"", nick);
+//	for (i = 0, j = strlen(namesbuf); nick[i] != 0; i++, j++)
+//		if (isspace(nick[i]))
+//			namesbuf[j] = '_';
+//		else
+//			namesbuf[j] = nick[i];
+//	namesbuf[j] = ' ';
+//	namesbuf[j+1] = 0;
 }
 
-transfer_t
-	*fnewtransfer(void *handle, const char *filename,
+transfer_t *fnewtransfer(void *handle, const char *filename,
 		const char *from, long size) {
 	transfer_t	*transfer;
 
@@ -1431,10 +1670,8 @@ transfer_t
 }
 
 void	fremove(transfer_t *transfer) {
-	free(transfer->from);
-	transfer->from = NULL;
-	free(transfer->remote);
-	transfer->remote = NULL;
+	FREESTR(transfer->from);
+	FREESTR(transfer->remote);
 	free(transfer);
 }
 
@@ -1515,9 +1752,9 @@ nFIRE_HANDLER(naim_file_progress) {
 	bwin->e.transfer->bytes = bytes;
 
 	if ((bwin->e.transfer->lastupdate+5) < now) {
-		window_echof(bwin, "STATUS %s/s, %lu/%lu (%i%%)\n",
+		window_echof(bwin, "STATUS %s/s, %lu/%lu (%.0f%%)\n",
 			dsize(bytes/(nowf-bwin->e.transfer->started)),
-			bytes, size, 100*bytes/size);
+			bytes, size, 100.0*bytes/size);
 		bwin->e.transfer->lastupdate = now;
 	}
 }
@@ -1661,7 +1898,7 @@ nFIRE_CTCPHAND(naim_ctcp_AUTOPEER) {
 					from);
 				blist = raddbuddy(conn, from, DEFAULT_GROUP, NULL);
 				bnewwin(conn, from, BUDDY);
-				firetalk_im_add_buddy(conn->conn, from, USER_GROUP(blist));
+				firetalk_im_add_buddy(conn->conn, from, USER_GROUP(blist), NULL);
 			} else {
 				if (getvar_int(conn, "autopeerverbose") > 0)
 					status_echof(conn, "Declining automatic negotiation with <font color=\"#00FFFF\">%s</font> (add <font color=\"#00FFFF\">%s</font> to your buddy list).\n", 
@@ -1810,12 +2047,14 @@ nFIRE_CTCPHAND(naim_ctcp_AUTOPEER) {
 nFIRE_CTCPHAND(naim_ctcp_default) {
 	conn_t	*conn = (conn_t *)client;
 
-	if (args == NULL)
-		echof(conn, "CTCP", "Unknown CTCP %s from <font color=\"#00FFFF\">%s</font>.\n",
-			command, from);
-	else
-		echof(conn, "CTCP", "Unknown CTCP %s from <font color=\"#00FFFF\">%s</font>: %s.\n",
-			command, from, args);
+	if (getvar_int(conn, "ctcpverbose") > 0) {
+		if (args == NULL)
+			echof(conn, "CTCP", "Unknown CTCP %s from <font color=\"#00FFFF\">%s</font>.\n",
+				command, from);
+		else
+			echof(conn, "CTCP", "Unknown CTCP %s from <font color=\"#00FFFF\">%s</font>: %s.\n",
+				command, from, args);
+	}
 }
 
 nFIRE_CTCPHAND(naim_ctcprep_VERSION) {
@@ -1916,7 +2155,7 @@ conn_t	*naim_newconn(int proto) {
 		conn->conn = firetalk_create_handle(proto, conn);
 
 		firetalk_register_callback(conn->conn, FC_DOINIT,
-			naim_budprof);
+			naim_doinit);
 		firetalk_register_callback(conn->conn, FC_POSTSELECT,
 			naim_postselect);
 		firetalk_register_callback(conn->conn, FC_CONNECTED,
@@ -1944,6 +2183,10 @@ conn_t	*naim_newconn(int proto) {
 			naim_im_handle);
 		firetalk_register_callback(conn->conn, FC_IM_GETACTION,
 			naim_act_handle);
+		firetalk_register_callback(conn->conn, FC_IM_BUDDYADDED,
+			naim_buddyadded);
+		firetalk_register_callback(conn->conn, FC_IM_BUDDYREMOVED,
+			naim_buddyremoved);
 		firetalk_register_callback(conn->conn, FC_IM_BUDDYONLINE,
 			naim_buddy_coming);
 		firetalk_register_callback(conn->conn, FC_IM_BUDDYOFFLINE,
@@ -1954,6 +2197,12 @@ conn_t	*naim_newconn(int proto) {
 			naim_buddy_unaway);
 		firetalk_register_callback(conn->conn, FC_IM_IDLEINFO,
 			naim_buddy_idle);
+		firetalk_register_callback(conn->conn, FC_IM_TYPINGINFO,
+			naim_buddy_typing);
+		firetalk_register_callback(conn->conn, FC_IM_EVILINFO,
+			naim_buddy_eviled);
+		firetalk_register_callback(conn->conn, FC_IM_CAPABILITIES,
+			naim_buddy_caps);
 		firetalk_register_callback(conn->conn, FC_IM_LISTBUDDY,
 			naim_buddylist);
 
