@@ -1,6 +1,6 @@
 /*  _ __   __ _ ___ __  __
 ** | '_ \ / _` |_ _|  \/  | naim
-** | | | | (_| || || |\/| | Copyright 1998-2003 Daniel Reed <n@ml.org>
+** | | | | (_| || || |\/| | Copyright 1998-2006 Daniel Reed <n@ml.org>
 ** |_| |_|\__,_|___|_|  |_| ncurses-based chat client
 */
 #include <naim/naim.h>
@@ -215,8 +215,12 @@ static int parsehtml_pair(const unsigned char *buf, int _pair, char *inbold, cha
 		buf += sizeof("rgb(")-1;
 		R = atoi(buf);
 		buf = strchr(buf, ',');
+		if (buf == NULL)
+			return(_pair);
 		G = atoi(buf);
 		buf = strchr(buf, ',');
+		if (buf == NULL)
+			return(_pair);
 		B = atoi(buf);
 	} else
 		return(_pair);
@@ -288,14 +292,15 @@ static unsigned long parsehtml_tag(h_t *h, unsigned char *text, int backup) {
 					t++;
 					while (isspace(*t))
 						t++;
-					if (*t == '"') {
-						t++;
-						while ((*t != 0) && (*t != '"')) {
+					if ((*t == '"') || (*t == '\'')) {
+						char	q = *t++;
+
+						while ((*t != 0) && (*t != q)) {
 							if (i < (sizeof(refbuf)-1))
 								refbuf[i++] = *t;
 							t++;
 						}
-						if (*t == '"')
+						if (*t == q)
 							t++;
 					} else {
 						while ((*t != 0) && (!isspace(*t))) {
@@ -318,10 +323,12 @@ static unsigned long parsehtml_tag(h_t *h, unsigned char *text, int backup) {
 						t++;
 						while (isspace(*t))
 							t++;
-						if (*t == '"') {
-							while ((*t != 0) && (*t != '"'))
+						if ((*t == '"') || (*t == '\'')) {
+							char	q = *t++;
+
+							while ((*t != 0) && (*t != q))
 								t++;
-							if (*t == '"')
+							if (*t == q)
 								t++;
 						} else
 							while ((*t != 0) && (!isspace(*t)))
@@ -343,7 +350,7 @@ static unsigned long parsehtml_tag(h_t *h, unsigned char *text, int backup) {
 		nw_wrap_addstr(h, "[IMAGE:");
 		nw_wrap_addstr(h, argbuf);
 		nw_wrap_addch(h, ']');
-	} else if CHECKTAG("BR") {
+	} else if (CHECKTAG("BR") || CHECKTAG("BR/")) {
 		nw_wrap_addstr(h, "\n ");
 		h->white = 1;
 	} else if CHECKTAG("HR") {
@@ -369,14 +376,15 @@ static unsigned long parsehtml_tag(h_t *h, unsigned char *text, int backup) {
 					t++;
 					while (isspace(*t))
 						t++;
-					if (*t == '"') {
-						t++;
-						while ((*t != 0) && (*t != '"')) {
+					if ((*t == '"') || (*t == '\'')) {
+						char	q = *t++;
+
+						while ((*t != 0) && (*t != q)) {
 							if (i < (sizeof(colbuf)-1))
 								colbuf[i++] = *t;
 							t++;
 						}
-						if (*t == '"')
+						if (*t == q)
 							t++;
 					} else {
 						while ((*t != 0) && (!isspace(*t))) {
@@ -408,10 +416,12 @@ static unsigned long parsehtml_tag(h_t *h, unsigned char *text, int backup) {
 						t++;
 						while (isspace(*t))
 							t++;
-						if (*t == '"') {
-							while ((*t != 0) && (*t != '"'))
+						if ((*t == '"') || (*t == '\'')) {
+							char	q = *t++;
+
+							while ((*t != 0) && (*t != q))
 								t++;
-							if (*t == '"')
+							if (*t == q)
 								t++;
 						} else
 							while ((*t != 0) && (!isspace(*t)))
@@ -453,14 +463,15 @@ static unsigned long parsehtml_tag(h_t *h, unsigned char *text, int backup) {
 					t++;
 					while (isspace(*t))
 						t++;
-					if (*t == '"') {
-						t++;
-						while ((*t != 0) && (*t != '"')) {
+					if ((*t == '"') || (*t == '\'')) {
+						char	q = *t++;
+
+						while ((*t != 0) && (*t != q)) {
 							if (i < (sizeof(colbuf)-1))
 								colbuf[i++] = *t;
 							t++;
 						}
-						if (*t == '"')
+						if (*t == q)
 							t++;
 					} else {
 						while ((*t != 0) && (!isspace(*t))) {
@@ -492,10 +503,12 @@ static unsigned long parsehtml_tag(h_t *h, unsigned char *text, int backup) {
 						t++;
 						while (isspace(*t))
 							t++;
-						if (*t == '"') {
-							while ((*t != 0) && (*t != '"'))
+						if ((*t == '"') || (*t == '\'')) {
+							char	q = *t++;
+
+							while ((*t != 0) && (*t != q))
 								t++;
-							if (*t == '"')
+							if (*t == q)
 								t++;
 						} else
 							while ((*t != 0) && (!isspace(*t)))
